'''
Bhuvan data GeoRSS to GeoJSON converter with metadata rescue
date: 27.10.18
author: Nikhil VJ
'''
print('\n\nGeoRSS Converter with metadata rescue')
print('Starting up the program, loading dependencies, please wait...\n\n')

import tornado.web
import tornado.ioloop
import json
import os
import time, datetime
import pandas as pd
from collections import OrderedDict
import webbrowser
import signal, sys # for catching Ctrl+C and exiting gracefully.

import feedparser
import geojson, geojson_utils
from bs4 import BeautifulSoup

# setting constants
root = os.path.dirname(__file__) # needed for tornado

outputFolder = 'output/'

# create folders if it doesn't exist
if not os.path.exists(outputFolder):
	os.makedirs(outputFolder)

print('Loaded dependencies, starting program.')

###########################
# Define all API call classes here

class upload(tornado.web.RequestHandler):
	def post(self):
		start = time.time()
		print('\nAPI call: upload()')
		payload = self.request.files['payload'][0]['body']
		filename = self.request.files['payload'][0]['filename'][:-4]

		print('payload: {}.xml . First 100 chars:'.format(filename))
		print(payload[:100])

		d = feedparser.parse(payload)
		print(len(d.entries),'features total.')

		collector = []
		count = 0
		for entry in d.entries:
			count += 1
			if count % 100 == 1: print(count,'shapes processed.')
			props = {'title': entry.title}
			
			# breaking up the HTML-i-fied description into properties metadata
			htmlBlock = entry.summary.replace('\xa0', "")
			# \xa0 : html whitespace char that was cropping up. So zapped it.
			soup =  BeautifulSoup(htmlBlock,'lxml')
			meta = soup.select('span')
			for n in range(int(len(meta)/2)):
				key = meta[n*2].text
				value = meta[n*2+1].text
				
				# eval to float or int if the string is number
				if value.lower().replace('.','',1).replace('e','',1).replace('-','',2).isdigit():
					if value.isdigit(): value = int(value)
					else: value = float(value)
				
				props[key] = value
			
			# print(props)
			
			# validation fix: if co-ords are in reversed order, then area is calculated negative
			# area.. from https://github.com/brandonxiang/geojson-python-utils
			if entry.where.type == 'Polygon':
				# only do for Polygon. 
				area = geojson_utils.area(entry.where)
				# print( area )
				if area < 0:
					# co-ordinates: reverse the order
					entry.where.coordinates[0].reverse()
					print('{}: after reversing, new area: {}'.format(count, geojson_utils.area(entry.where)) )
			
			collector.append( geojson.Feature(geometry=entry.where, properties=props, id=count) )

		with open(outputFolder+filename+'.geojson','w') as f:
			f.write( geojson.dumps(geojson.FeatureCollection(collector), indent=2 ) )	
			# f.write( geojson.dumps(geojson.FeatureCollection(collector) ) )

		# second run : make CSV
		dfCollector = []
		for x in collector:
			row = OrderedDict({'id':x.id})
			row.update(x.properties)
			row['geometry_type'] = x.geometry.type
			row['geometry_coordinates'] = geojson.dumps(x.geometry.coordinates)
			# Extra: if it's point coords, then make in lat and long columns so that the CSV is easily mappable in geojson.io, google my maps etc
			if row['geometry_type'] == 'Point':
				row['latitude'] = x.geometry.coordinates[1]
				row['longitude'] = x.geometry.coordinates[0]
			
			dfCollector.append(row)

		df = pd.DataFrame(dfCollector).set_index('id')
		df.to_csv(outputFolder+filename+'.csv')

		# meta fields
		metaFields = [ x for x in list(df.columns) if x not in ['title','geometry_type','geometry_coordinates','latitude','longitude']]
		# geometry types
		shape_types = df['geometry_type'].unique().tolist()

		returnMessage = """GeoJSON: <a href="{}{}.geojson">{}.geojson</a><br>
		CSV version: <a href="{}{}.csv">{}.csv</a> (for inspecting the metadata)<br>
		Total {} shapes of type(s): {}<br>
		Meta fields: {}
		""".format(outputFolder,filename,filename,outputFolder,filename,filename, count, ', '.join(shape_types), ', '.join(metaFields))
		
		self.write(returnMessage)
		end = time.time()
		print("upload POST call took {} seconds.".format(round(end-start,2)))


###################
# List API calls and the class they will be directed to, here
def make_app():
	return tornado.web.Application([
		#(r"/API/data", APIHandler),
		(r"/API/upload", upload),
		(r"/(.*)", tornado.web.StaticFileHandler, {"path": root, "default_filename": "index.html"})
	])
###################
# Done with API calls defining. 

####################
# Functions

# for catching Ctrl+C and exiting gracefully. From https://nattster.wordpress.com/2013/06/05/catch-kill-signal-in-python/
def signal_term_handler(signal, frame):
	# to do: Make this work in windows, ra!
	print('\nClosing Program.\n')
	sys.exit(0)

###################################################			
# Now, at last, the __main__ function sets up the tornado server

if __name__ == "__main__":
	signal.signal(signal.SIGINT, signal_term_handler)
	app = make_app()
	portnum = 5000
	while True: # loop to increment the port number till we find one that isn't occupied
		try:
			port = int(os.environ.get("PORT", portnum))
			app.listen(port)
			break
		except OSError:
			portnum += 1
			if portnum > 9999: 
				print('Thats it I give up! From 5000 to 9999 not a single port number is free! Can\'t run your server, sorry!')
				sys.exit()

	thisURL = "http://localhost:" + str(port)
	webbrowser.open(thisURL)
	print("Open " + thisURL + " in your Web Browser if you don't see it opening automatically within 5 seconds.")
	tornado.ioloop.IOLoop.current().start()

